import logging
import threading
from datetime import datetime
from queue import Queue, Empty
from pymongo import MongoClient, ASCENDING, DESCENDING

from blocks_genesis._auth.blocks_context import BlocksContextManager
from blocks_genesis._core.secret_loader import get_blocks_secret
from blocks_genesis._lmt.activity import Activity


class MongoBatchLogger:
    def __init__(self, batch_size=50, flush_interval_sec=2.0):
        self.batch_size = batch_size
        self.flush_interval_sec = flush_interval_sec
        self.blocks_secret = get_blocks_secret()
        # Lazy initialization of MongoDB connection
        mongo_client = MongoClient(self.blocks_secret.LogConnectionString)
        db = mongo_client[self.blocks_secret.LogDatabaseName]

        if self.blocks_secret.ServiceName not in db.list_collection_names():
            db.create_collection(
                self.blocks_secret.ServiceName,
                timeseries={
                    "timeField": "Timestamp",
                    "metaField": "TenantId",
                    "granularity": "minutes"
                }
            )
            db[self.blocks_secret.ServiceName].create_index(
                [("TenantId", ASCENDING), ("Timestamp", DESCENDING)],
                name="Tenant_Timestamp_Index"
            )

        self.collection = db[self.blocks_secret.ServiceName]
        self.queue = Queue()
        self._stop_event = threading.Event()
        self.worker_thread = threading.Thread(target=self._background_worker, daemon=True)
        self.worker_thread.start()

    def enqueue(self, record: logging.LogRecord):
        doc = {
            "Timestamp": datetime.now(),
            "Level": record.levelname,
            "Message": record.getMessage(),
            "TenantId": record.TenantId or "miscellaneous",
            "LoggerName": record.name,
            "TraceId": record.TraceId or Activity.get_trace_id(),
            "SpanId": record.SpanId or Activity.get_span_id(),
        }
        self.queue.put(doc)

    def _background_worker(self):
        batch = []
        while not self._stop_event.is_set():
            try:
                doc = self.queue.get(timeout=self.flush_interval_sec)
                batch.append(doc)
            except Empty:
                pass

            if batch and (len(batch) >= self.batch_size or self._stop_event.is_set()):
                try:
                    self.collection.insert_many(batch)
                except Exception as e:
                    print(f"[MongoBatchLogger] Insert error: {e}")
                batch.clear()

        # flush remaining logs on shutdown
        if batch:
            try:
                self.collection.insert_many(batch)
            except Exception as e:
                print(f"[MongoBatchLogger] Insert error on shutdown: {e}")

    def stop(self):
        self._stop_event.set()
        self.worker_thread.join()


class MongoHandler(logging.Handler):
    _mongo_logger = None

    def __init__(self, batch_size=50, flush_interval_sec=2.0):
        super().__init__()
        if not MongoHandler._mongo_logger:
            MongoHandler._mongo_logger = MongoBatchLogger(batch_size, flush_interval_sec)
        self.mongo_logger = MongoHandler._mongo_logger

    def emit(self, record: logging.LogRecord):
        try:
            self.mongo_logger.enqueue(record)
        except Exception:
            self.handleError(record)


class TraceContextFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        """Add trace context to log records."""
        record.TenantId = BlocksContextManager.get_context().tenant_id if BlocksContextManager.get_context() else "miscellaneous"
        print(f"[TraceContextFilter] TenantId: {record.TenantId}")
        record.TraceId = Activity.get_trace_id()
        record.SpanId = Activity.get_span_id()
        return True



from dataclasses import dataclass
from typing import Any, Optional

@dataclass
class ConsumerMessage:
    consumer_name: str
    payload: Any  
    context: Optional[str] = None
    routing_key: str = ""




from blocks_genesis._message.event_registry import get_event_handler

class Consumer:
    async def process_message(self, type: str, body: dict):
        handler = get_event_handler(type)

        if callable(handler):  # If it’s a function
            await handler(body)
        elif hasattr(handler, "handle"):  # If it's a class with `handle`
            instance = handler()
            await instance.handle(body)
        else:
            raise TypeError(f"Handler for type '{type}' is not callable or doesn't implement `handle()`")



from dataclasses import dataclass, field
from typing import List, Optional
from datetime import timedelta


@dataclass
class AzureServiceBusConfiguration:
    queues: List[str] = field(default_factory=list)
    topics: List[str] = field(default_factory=list)
    queue_max_size_in_megabytes: int = 1024
    queue_max_delivery_count: int = 2
    queue_prefetch_count: int = 10
    queue_default_message_time_to_live: timedelta = timedelta(days=7)

    topic_prefetch_count: int = 10
    topic_max_size_in_megabytes: int = 1024
    topic_default_message_time_to_live: timedelta = timedelta(days=30)

    topic_subscription_max_delivery_count: int = 2
    topic_subscription_default_message_time_to_live: timedelta = timedelta(days=7)
    max_concurrent_calls: int = 5
    max_message_processing_time_in_minutes: int = 60
    message_lock_renewal_interval_seconds: int = 270

    def set_queues(self, queue_list: List[str]):
        self.queues = [q.lower() for q in queue_list if q and q.strip()]

    def set_topics(self, topic_list: List[str]):
        self.topics = [t.lower() for t in topic_list if t and t.strip()]


@dataclass
class MessageConfiguration:
    connection: Optional[str] = None
    service_name: Optional[str] = None
    azure_service_bus_configuration: Optional[AzureServiceBusConfiguration] = None

    def get_subscription_name(self, topic_name: str) -> str:
        return f"{topic_name}_sub_{self.service_name}"



# tracing.py
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from blocks_genesis._lmt.mongo_trace_exporter import MongoDBTraceExporter
from blocks_genesis._core.secret_loader import get_blocks_secret

def configure_tracing() -> None:
    """Configures OpenTelemetry tracing for the FastAPI application."""
    trace.set_tracer_provider(
        TracerProvider(
            resource=Resource.create({SERVICE_NAME: get_blocks_secret().ServiceName})
        )
    )

    exporter = MongoDBTraceExporter()

    processor = BatchSpanProcessor(exporter)
    trace.get_tracer_provider().add_span_processor(processor)




from blocks_genesis._auth.blocks_context import BlocksContext, BlocksContextManager
from blocks_genesis._cache import CacheClient
from blocks_genesis._cache.cache_provider import CacheProvider
from blocks_genesis._database.db_context_provider import DbContextProvider
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._tenant.tenant import Tenant
from blocks_genesis._tenant.tenant_service import TenantService, get_tenant_service
from blocks_genesis._message.azure.azure_message_client import AzureMessageClient
from blocks_genesis._core.api import close_lifespan, configure_lifespan, configure_middlewares
from blocks_genesis._core.worker import WorkerConsoleApp
from blocks_genesis._core.configuration import get_configurations, load_configurations
from blocks_genesis._entities.base_entity import BaseEntity
from blocks_genesis._lmt.activity import Activity
from blocks_genesis._message.consumer_message import ConsumerMessage
from blocks_genesis._message.message_client import MessageClient
from blocks_genesis._utilities.crypto_service import CryptoService
from blocks_genesis._auth.auth import authorize
from blocks_genesis._message.message_configuration import AzureServiceBusConfiguration, MessageConfiguration

__all__ = [
    "BlocksContext",
    "BlocksContextManager",
    "CacheClient",
    "CacheProvider",
    "DbContextProvider",
    "DbContext",
    "Tenant",
    "TenantService",
    "get_tenant_service",
    "AzureMessageClient",
    "MessageConfiguration",
    "close_lifespan",
    "configure_lifespan",
    "configure_middlewares",
    "WorkerConsoleApp",
    "get_configurations",
    "load_configurations",
    "BaseEntity",
    "Activity",
    "ConsumerMessage",
    "MessageClient",
    "CryptoService",
    "AzureServiceBusConfiguration",
    "authorize"
]



from typing import Optional
from blocks_genesis._database.db_context_provider import DbContextProvider


class DbContext:
    _provider: Optional[DbContextProvider] = None

    @classmethod
    def set_provider(cls, provider: DbContextProvider) -> None:
        cls._provider = provider

    @classmethod
    def get_provider(cls) -> DbContextProvider:
        if cls._provider is None:
            raise RuntimeError("No DbContextProvider registered.")
        return cls._provider

    @classmethod
    def clear(cls) -> None:
        cls._provider = None



import logging
from typing import Dict, Optional
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.database import Database
from pymongo.monitoring import register
from contextvars import ContextVar

from blocks_genesis._auth.blocks_context import BlocksContext, BlocksContextManager
from blocks_genesis._database.mongo_event_subscriber import MongoEventSubscriber
from blocks_genesis._tenant.tenant_service import get_tenant_service
from blocks_genesis._database.db_context_provider import DbContextProvider

_db_cache: ContextVar[Dict[str, Database]] = ContextVar("_db_cache", default={})
_client_cache: ContextVar[Dict[str, MongoClient]] = ContextVar("_client_cache", default={})

_logger = logging.getLogger(__name__)


class MongoDbContextProvider(DbContextProvider):
    def __init__(self):
        self._logger = _logger
        self._tenants = get_tenant_service()
        register(MongoEventSubscriber())

    async def get_database(self, tenant_id: Optional[str] = None) -> Optional[Database]:
        tenant_id = tenant_id or getattr(BlocksContextManager.get_context(), 'tenant_id', None)
        if not tenant_id:
            self._logger.warning("Tenant ID is missing in context")
            return None

        dbs = _db_cache.get()
        clients = _client_cache.get()

        if tenant_id in dbs:
            return dbs[tenant_id]

        db_name, connection_string = await self._tenants.get_db_connection(tenant_id)
        if not connection_string or not db_name:
            raise ValueError(f"Missing connection info for tenant {tenant_id}")

        if connection_string not in clients:
            clients[connection_string] = self._create_mongo_client(connection_string)

        db = clients[connection_string][db_name]

        # Update context vars with new copy
        new_dbs = dict(dbs)
        new_dbs[tenant_id] = db
        _db_cache.set(new_dbs)
        _client_cache.set(dict(clients))

        return db

    def get_database_by_connection(self, connection_string: str, database_name: str) -> Database:
        if not connection_string:
            raise ValueError("Connection string cannot be empty or None.")
        if not database_name:
            raise ValueError("Database name cannot be empty or None.")

        db_key = database_name.lower()
        dbs = _db_cache.get()
        clients = _client_cache.get()

        if db_key in dbs:
            return dbs[db_key]

        if connection_string not in clients:
            clients[connection_string] = self._create_mongo_client(connection_string)

        db = clients[connection_string][database_name]

        new_dbs = dict(dbs)
        new_dbs[db_key] = db
        _db_cache.set(new_dbs)
        _client_cache.set(dict(clients))

        return db

    async def get_collection(self, collection_name: str, tenant_id: Optional[str] = None) -> Collection:
        db = await self.get_database(tenant_id)
        if not db:
            raise RuntimeError("No database found for tenant")
        return db[collection_name]

    def _create_mongo_client(self, connection_string: str) -> MongoClient:
        self._logger.info("Creating new MongoClient for connection string.")
        return MongoClient(connection_string)



# event_registry.py

_event_handlers = {}

def register_event_handler(event_type: str):
    def wrapper(cls_or_func):
        _event_handlers[event_type] = cls_or_func
        return cls_or_func
    return wrapper

def get_event_handler(event_type: str):
    handler = _event_handlers.get(event_type)
    if not handler:
        raise ValueError(f"No handler registered for event type: {event_type}")
    return handler



import asyncio
import logging
import sys

from blocks_genesis._core.worker import WorkerConsoleApp
from blocks_genesis._message.message_configuration import AzureServiceBusConfiguration, MessageConfiguration


logger = logging.getLogger(__name__)
message_config = MessageConfiguration(
    azure_service_bus_configuration=AzureServiceBusConfiguration(
        queues=["ai_queue", "ai_queue_2nd"],
        topics=[]
    )
)


def main():
    app = WorkerConsoleApp("blocks_ai_worker", message_config)

    try:
        asyncio.run(app.run())
    except KeyboardInterrupt:
        logger.exception("\n👋 Graceful shutdown by user")
    except Exception as e:
        logger.exception(f"💥 Fatal error: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import asyncio
from contextlib import asynccontextmanager
import logging

from blocks_genesis._cache.cache_provider import CacheProvider
from blocks_genesis._cache.redis_client import RedisClient
from blocks_genesis._core.secret_loader import SecretLoader
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._database.mongo_context import MongoDbContextProvider
from blocks_genesis._message.azure.azure_message_client import AzureMessageClient
from blocks_genesis._message.azure.azure_message_worker import AzureMessageWorker
from blocks_genesis._message.azure.config_azure_service_bus import ConfigAzureServiceBus
from blocks_genesis._message.message_configuration import MessageConfiguration
from blocks_genesis._lmt.log_config import configure_logger
from blocks_genesis._lmt.mongo_log_exporter import MongoHandler
from blocks_genesis._lmt.tracing import configure_tracing
from blocks_genesis._tenant.tenant_service import initialize_tenant_service



class WorkerConsoleApp:
    def __init__(self, name: str, message_config: MessageConfiguration):
        self.message_worker: AzureMessageWorker = None
        self.name = name
        self.logger = logging.getLogger(__name__)
        self.message_config = message_config

    @asynccontextmanager
    async def setup_services(self):
        self.logger.info("🚀 Starting Blocks AI Worker Console App...")

        try:
            self.logger.info("🔐 Loading secrets...")
            await SecretLoader(self.name).load_secrets()
            self.logger.info("✅ Secrets loaded successfully")

            configure_logger()
            self.logger.info("📝 Logger configured")

            configure_tracing()
            self.logger.info("🔍 Tracing configured")

            CacheProvider.set_client(RedisClient())
            await initialize_tenant_service()
            DbContext.set_provider(MongoDbContextProvider())
            self.logger.info("✅ Cache, TenantService, and Mongo Context initialized")

            

            ConfigAzureServiceBus().configure_queue_and_topic(self.message_config)
            AzureMessageClient.initialize(self.message_config)

            self.message_worker = AzureMessageWorker(self.message_config)
            self.message_worker.initialize()

            self.logger.info("✅ Azure Message Worker initialized and ready")
            yield self.message_worker

        except Exception as ex:
            self.logger.error(f"❌ Startup failed: {ex}", exc_info=True)
            raise

        finally:
            await self.cleanup()

    async def cleanup(self):
        self.logger.info("🛑 Cleaning up services...")

        if self.message_worker:
            await self.message_worker.stop()

        if hasattr(MongoHandler, '_mongo_logger') and MongoHandler._mongo_logger:
            MongoHandler._mongo_logger.stop()

        self.logger.info("✅ Shutdown complete")

    async def run(self):
        async with self.setup_services() as worker:
            self.logger.info("🔄 Worker running... Press Ctrl+C to stop")
            try:
                await worker.run()
            except asyncio.CancelledError:
                self.logger.info("🛑 Received cancellation signal")
            except KeyboardInterrupt:
                self.logger.info("⏹️ Received interrupt signal")







from dataclasses import dataclass
import logging
from contextlib import asynccontextmanager
from pathlib import Path
from fastapi import FastAPI
from blocks_genesis._auth.auth import authorize
from blocks_genesis._core.api import close_lifespan, configure_lifespan, configure_middlewares
from blocks_genesis._core.configuration import get_configurations, load_configurations
from blocks_genesis._database.db_context import DbContext
from blocks_genesis._message.azure.azure_message_client import AzureMessageClient
from blocks_genesis._message.consumer_message import ConsumerMessage
from blocks_genesis._message.message_configuration import AzureServiceBusConfiguration, MessageConfiguration



logger = logging.getLogger(__name__)
message_config = MessageConfiguration(
    azure_service_bus_configuration=AzureServiceBusConfiguration(
        queues=["ai_queue"],
        topics=[]
    )
)

config_dir = Path(__file__).resolve().parent / "config"
print(config_dir)
load_configurations(config_dir)
config = get_configurations()

@asynccontextmanager
async def lifespan(app: FastAPI):
    await configure_lifespan("blocks_ai_api", message_config)
    logger.info("✅ All services initialized!")

    yield  # app running here

    await close_lifespan()
    logger.info("🛑 App shutting down...")



app = FastAPI(lifespan=lifespan, debug=True)

# Add middleware in order
configure_middlewares(app, config["is_local"])






@app.get("/")
async def root():
    logger.info("Root endpoint accessed")
    client = AzureMessageClient.get_instance()
    await client.send_to_consumer_async(ConsumerMessage(
        consumer_name="ai_queue",
        payload= AiMessage("Hello from AI API!"),
    ))
    db = DbContext.get_provider()
    return {"message": "Hello World", "secrets_loaded": True}



@app.get("/health", dependencies=[authorize(bypass_authorization=True)])
async def health():
    return {
        "status": "healthy",
        "secrets_status": "loaded" ,
    }
    
  
    
    
@dataclass
class AiMessage:
    message: str




from pathlib import Path
import json
import os

_app_configurations = None

def load_configurations(config_dir: str, env: str = None):
    global _app_configurations 
    env = env or os.getenv("APP_ENV", "dev")
    path = Path(config_dir) / f"{env}.json"
    if not path.exists():
        raise FileNotFoundError(f"Config file not found: {path}")
    with open(path) as f:
        _app_configurations = json.load(f)  

def get_configurations():
    if _app_configurations is None:
        raise Exception("Configurations not loaded")
    return _app_configurations
